#include<linux/init.h>                  // Basic Initialization
#include<linux/module.h>                // Needed for all the modules.
#include<linux/kernel.h>                // Allow to use the printk function.
#include<linux/device.h>                // The device creation API's are defined here.
#include<linux/cdev.h>                  // The cdev structure is defined in it.
#include<linux/fs.h>                    // To register the device. 

MODULE_LICENSE("GPL");                                  	 // This is a GPL-licesed module.        
MODULE_AUTHOR("FOSSEE TEAM");                           	 // Macro used to give author name to a module.
MODULE_DESCRIPTION("Create A New Character Device.\n");          // Macro gives short description of a module.

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////// 

static dev_t device_num;         // Global variable that holds the device number.
static struct class *cl;         // structure creating the class for a device.  

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* cdev structure for a character device .*/

static struct cdev my_cdev;    

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The prototypes of the file operations such as open and release that implemented in this driver */

static int mydevice_open (struct inode *inod, struct file *filp);
static int mydevice_release (struct inode *inod, struct file *filp);
  
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The file_operation structure of a device.*/

static struct file_operations fops=
{
.open=mydevice_open,
.release=mydevice_release,

};


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* Initialization function */

static int init_function(void)
{
	printk(KERN_INFO "simple_driver : Inside the init function of a driver.\n");


/********************************************************************************************************************
int alloc_chrdev_region(dev_t *dev, unsigned int firstminor, unsigned int count, char *name) 

Description : It is use for the dynamic allocation of a device number by the kernel.
*********************************************************************************************************************/

	if(alloc_chrdev_region(&device_num, 0, 1, "my_device") < 0)
	{	
		printk(KERN_INFO "simple_driver : The device number allocation fails.\n");
		return -1;
	}

/******************************************************************************************************************** 
The important macros are as follows : 
1. MAJOR(dev_t dev) 		:	Extracts the major number from device number.
2. MINOR(dev_t dev)     	:	Extracts the minor number from device number.
*********************************************************************************************************************/
	
	printk(KERN_INFO "simple_driver : The device number allocation done successfully.\n");	
	printk(KERN_INFO "simple_driver : Major and minor numbers of my_device are : %d, %d\n", MAJOR(device_num), 		  MINOR(device_num));

/******************************************************************************************************************** 
struct class *class_create(struct module *owner, const char *name)

Description - This function use to create a new class under sysfs file system.
*********************************************************************************************************************/
	
	if((cl = class_create(THIS_MODULE,"my_class")) == NULL)
	{
		
		printk(KERN_INFO "simple_driver : the class creation fails here.\n");	
		unregister_chrdev_region(device_num,1);
		return -1;
	}
	else
	{
		printk(KERN_INFO "simple_driver : the class is created successfully here.\n");

	}

/******************************************************************************************************************** 
struct device *device_create(struct class *class, struct device *parent, dev_t devt, void *drvdata,  const char *fmt,.)

Description - It creates a device and registers it with sysfs and creates the device node under dev directory.
*********************************************************************************************************************/
	
	if(device_create(cl,NULL,device_num,NULL,"new_device") == NULL)
	{
		printk(KERN_INFO "simple_driver : the device node creation fails here.\n");

/******************************************************************************************************************** 
void class_destroy (struct class *class)

Description - It destroys a struct class structure.
*********************************************************************************************************************/		
		class_destroy(cl);
		unregister_chrdev_region(device_num,1);
		return -1;
	}
	else 
        {
		printk(KERN_INFO "simple_driver : the device node is created successfully.\n");
	
	}

/******************************************************************************************************************** 
void cdev_init (struct cdev *cdev, const struct file_operations *fops)	
Description — initialize a cdev structure 

int cdev_add (struct cdev *p, dev_t dev, unsigned count)	
Description — add a char device to the system 
*********************************************************************************************************************/		
	cdev_init(&my_cdev,&fops);
	if(cdev_add(&my_cdev,device_num,1) == -1)
	{
		device_destroy(cl,device_num);
		class_destroy(cl);
			   unregister_chrdev_region(device_num,1);
		return -1;
	}
	else
	{
		printk(KERN_INFO "simple_driver : the cdev_structure initialization done successfully.\n");
			
	}
	return(0);

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* the open function of a device */

static int mydevice_open (struct inode *inod, struct file *filp)
{
	printk(KERN_INFO "simple_driver : The new_device opened successfully.\n");

	return 0;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The close function of a device */

static int mydevice_release(struct inode *inod, struct file *filp)
{
	printk(KERN_INFO "simple_driver : the new_device is closed successfully.\n");
	
	return(0);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The cleanup function */

static void exit_function(void)
{

	printk(KERN_INFO "simple_driver : Inside the exit function of the driver.\n");

/******************************************************************************************************************** 
void cdev_del(struct cdev *cdev)

Description - This function removes the cdev structure from the system.
*********************************************************************************************************************/		
	
	cdev_del(&my_cdev);
	printk(KERN_INFO "simple_driver : The cdev structure removed from the system.\n");
	

/******************************************************************************************************************** 
void device_destroy(struct class *class,dev_t devt).

Description - This function unregisters and cleans up a device that was created with a call to device_create. 
*********************************************************************************************************************/		
	device_destroy(cl,device_num);
	printk(KERN_INFO "simple_driver : The device removed from the /dev directory and the system.\n");

	class_destroy(cl);
	printk(KERN_INFO "simple_driver : The class of the device removed from the system.\n");
	
/*******************************************************************************************************************
void unregister_chrdev_region(dev_t first,unsigned int count).
Description - It is used to free our device number.
	
********************************************************************************************************************/
	unregister_chrdev_region(device_num, 1);
	printk(KERN_INFO "simple_driver : The device is unregistered successfully.\n");

}

module_init(init_function);
module_exit(exit_function);



